import logging
from datetime import datetime
from ...services.locale.Locale import LocaleData
from statistics import mean
_LOGGER = logging.getLogger(__name__)

EXPORT_FACTOR = 0.9

class CurrentPeak:
    def __init__(self, data_type: type, initval, startpeaks:dict, locale: LocaleData, options_use_history: bool = False, mock_dt: datetime | None = None):
        self._options_peaks: dict = startpeaks
        self._value = initval
        self._history: dict[str, list[float|int]] = {}
        self._locale: LocaleData = locale
        self._peaks: list[float|int] = []
        self._active: bool = options_use_history
        self.mock_dt: datetime | None = mock_dt
        
    @property
    def dt(self) -> datetime:
        return self.mock_dt if self.mock_dt is not None else datetime.now()

    @dt.setter
    def dt(self, val: datetime):
        self.mock_dt = val

    @property
    def value(self): # type: ignore
        self._refresh_peak()
        return self._value

    @value.setter
    def value(self, val): # pylint:disable=invalid-overridden-method
        self._value = val
        self.update_history([val] if not isinstance(val, list) else val)

    @property
    def history(self) -> dict:
        return self._history

    def _make_key(self, year = None) -> str:
        if year is None:
            year = self.dt.year
        return f"{str(year)}_{str(self.dt.month)}"

    async def async_update(self, peaks: list) -> None:
        self.update_history(peaks)

    # def update_history(self, peaks: list) -> None:
    #     self._peaks.extend(peaks) if isinstance(peaks, list) else self._peaks.append(peaks)
    #     current_key = self._make_key()
    #     if current_key not in self._history:
    #         self._history[current_key] = []
    #     self._history[current_key].extend(self._peaks)
    #     self._refresh_peak()

    def update_history(self, peaks: list) -> None:	
        _key = self._make_key()	
        self._history[_key] = peaks	
        self._refresh_peak()

    def _refresh_peak(self) -> None:
        cc = self._locale.data.query_model.get_currently_obeserved_peak(self.dt)
        if len(self._peaks) and cc > min(self._peaks):
            self._value = self._locale.data.query_model.observed_peak
        else:
            self._value = self._get_peak()

    def _get_peak(self) -> float:
        try:
            options_start = self._options_peaks.get(self.dt.month, self._options_peaks.get(str(self.dt.month), 0))
            past_key = self._make_key(self.dt.year - 1)
            current_key = self._make_key()
            past_mean = mean(self._history.get(past_key, [0]))
            current_mean = mean(self._history.get(current_key, [0]))
            max_mean = max(past_mean, current_mean)
            if max_mean == past_mean and past_mean > options_start:
                return min(self._history[past_key]) * EXPORT_FACTOR
            elif max_mean == current_mean and current_mean > options_start:
                return min(self._history[current_key])
            else:
                return options_start
        except Exception as e:
            _LOGGER.error(f"Error in get_peak: {e}")
            return 0

    def import_from_service(self, importdto: dict, current:bool = False) -> dict:
        """Import the dict passed from service or on loading hass"""
        ret = {}
        validation_errors = []
        for month in importdto.keys():
            if all([
                self._validate_month(month, validation_errors),
                self._validate_value(month, importdto[month], validation_errors)
            ]):
                self._history[month] = self._set_history_value(importdto[month])
        if len(validation_errors):
            ret["status"] = "Error"
            ret["errors"] = validation_errors
        else:
            ret["status"] = "OK"
            ret["errors"] = ["No errors"]
        self._value = self._get_peak()
        return ret

    def _set_history_value(self, value: float|int|list) -> list:
        if isinstance(value,float|int):
            return [value]
        if isinstance(value, list):
            return value

    def _validate_value(self, monthkey: str, value: any, validation_errors: list) -> bool: # type: ignore
        """check if value is valid"""
        if isinstance(value, list):
            for _value in value:
                if not self._validate_value(monthkey, _value, validation_errors):
                    return False
            return True
        try:
            _value = float(value)
            if _value < 0:
                validation_errors.append(f"Invalid value: {monthkey} - {value}")
                return False
            return True
        except Exception as e:
            validation_errors.append(f"Error with value ({e}): {monthkey} - {value}")
            return False

    def _validate_month(self, monthkey: str, validation_errors: list) -> bool:
        """check if monthkey is valid"""
        _components = monthkey.split("_")
        if len(_components) != 2:
            validation_errors.append(f"Invalid monthkey: {monthkey}")
            return False
        try:
            _year = int(_components[0])
            _month = int(_components[1])
            if any(
                [
                    _year > self.dt.year, 
                    _year == self.dt.year and _month > self.dt.month
                ]):
                validation_errors.append(f"Period is in the future. Cannot add to history: {monthkey}")
                return False
            if _month > 12:
                validation_errors.append(f"Invalid month: {monthkey}")
                return False
            return True
        except Exception as e:
            validation_errors.append(f"Error with month ({e}): {monthkey}")
            return False



from datetime import datetime
from abc import abstractmethod
from .querytypes.queryservice import QueryService
from ...models.locale.peaks_model import PeaksModel
from ...models.locale.sumcounter import SumCounter
from ...models.locale.price.locale_price import LocalePrice

class ILocaleQuery:

    @property
    @abstractmethod
    def dt(self) -> datetime:
        pass


    @abstractmethod
    def set_mock_dt(self, val: datetime | None):
        pass

    @abstractmethod
    def reset(self) -> None:
        pass

    @property
    @abstractmethod
    def peaks(self) -> PeaksModel:
        pass

    @property
    @abstractmethod
    def sum_counter(self) -> SumCounter:
        pass

    @property
    @abstractmethod
    def charged_peak(self) -> float:
        pass

    @property
    @abstractmethod
    def observed_peak(self) -> float:
        pass

    @property
    @abstractmethod
    def price(self) -> LocalePrice:
        pass
    
    @price.setter
    @abstractmethod
    def price(self, value: LocalePrice):
        pass

    @abstractmethod
    def _sanitize_values(self):
        pass

    @abstractmethod
    async def async_reset(self) -> None:
        pass

    @abstractmethod
    async def async_set_query_service(self, service: QueryService) -> None:
        pass

    @abstractmethod
    async def async_try_update(self, new_val, timestamp: datetime | None = None):
        pass

    @abstractmethod
    async def async_set_update_for_groupby(self, new_val, dt):
        pass

    @abstractmethod
    async def async_update_peaks(self):
        pass

    @abstractmethod
    async def async_reset_values(self, new_val, dt=datetime.now()):
        pass

    @abstractmethod
    async def async_sanitize_values(self):
        pass

    @abstractmethod
    def get_currently_obeserved_peak(self, timestamp: datetime = datetime.now()) -> float:
        pass


import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from statistics import mean, stdev

_LOGGER = logging.getLogger(__name__)


@dataclass
class ISpotPriceDTO:
    state: float = 0
    today: list = field(default_factory=lambda: [])
    tomorrow: list = field(default_factory=lambda: [])
    average: float = 0
    stdev: float = 0
    currency: str = ""
    price_in_cent: bool = False #todo: add support when source exposes this option.
    tomorrow_valid: bool = False

    async def set_model(self, ret):
        try:
            self.today = list(ret.attributes.get("today"))
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices. Unsolveable error. {e}"
            )
            return
        self.tomorrow_valid = bool(ret.attributes.get("tomorrow_valid", False))
        _tomorrow = ret.attributes.get("tomorrow", [])
        if _tomorrow is not None:
            self.tomorrow = list(_tomorrow)
        else:
            self.tomorrow = []
        self.currency = str(ret.attributes.get("currency", ""))
        self.state = ret.state
        self.average = self._set_average(ret)
        self.stdev = stdev(self.today) if len(self.today) > 1 else 0
        self.price_in_cent = self._set_price_in_cent(ret)

    @abstractmethod
    def _set_price_in_cent(self, ret) -> bool:
        pass

    @abstractmethod
    def _set_average(self, ret) -> float:
        pass


@dataclass
class EnergiDataServiceDTO(ISpotPriceDTO):

    def _set_price_in_cent(self, ret) -> bool:
            return bool(ret.attributes.get("use_cent", False))

    def _set_average(self, ret) -> float:
        try:
            return round(mean(self.today), 2)
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices from EnergiDataService. Unsolveable error. {e}"
            )
            return 0



@dataclass
class NordpoolDTO(ISpotPriceDTO):
    
    def _set_price_in_cent(self, ret) -> bool:
        return bool(ret.attributes.get("price_in_cent", False))

    def _set_average(self, ret) -> float:
        try:
            return float(str(ret.attributes.get("average", 0)))
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices from Nordpool. Unsolveable error. {e}"
            )
            return 0


WEEKDAYBITS = {
    1: 1,
    2: 2,
    3: 4,
    4: 8,
    5: 16,
    6: 32,
    0: 64
}

class WeekdayTranslation:
    @staticmethod
    def get_days(inp: int) -> list:
        _rest = inp
        ret = []
        while _rest > 0:
            val = max({k: v for k, v in WEEKDAYBITS.items() if v <= _rest}.values())
            ret.append([k for k, v in WEEKDAYBITS.items() if v == val][0])
            _rest = _rest - val
        return ret

    @staticmethod
    def set_days(input: list) -> int:
        ret = 0
        for i in input:
            ret += WEEKDAYBITS[i]
        return ret


# testdays = [1,3,5]
# schema = WeekdayTranslation.set_days(testdays)
# print(schema)
# print(WeekdayTranslation.get_days(schema))




from dataclasses import dataclass
from datetime import datetime
import logging
import time

_LOGGER = logging.getLogger(__name__)


@dataclass
class Day:
    sessions: int
    charge: float


class EnergyWeekly:
    def __init__(self):
        self.model: dict[int, Day] = {}

    async def async_setup(self, incoming_dict: dict | None = None):
        if incoming_dict is not None:
            await self.async_unpack(incoming_dict)
        else:
            await self.async_set_init_model()

    async def async_set_init_model(self):
        ret = {}
        for i in range(0, 7):
            ret[i] = Day(0, 0)
        self.model = ret

    @property
    def export(self) -> dict[int, dict[str, float]]:
        "export the model to HA-friendly states"
        ret = {}
        for idx, m in self.model.items():
            m_ret = {"sessions": m.sessions, "total_charge": m.charge}
            ret[idx] = m_ret
        return ret

    @property
    def average(self) -> float:
        sessions = sum([h.sessions for h in self.model.values()])
        charge = sum([h.charge for h in self.model.values()])
        ret = 0.0
        try:
            ret = charge / sessions
        except:
            pass
            #_LOGGER.debug(f"Could not calculate average")
        return round(ret, 1)

    @property
    def total_sessions(self) -> int:
        try:
            return sum([h.sessions for h in self.model.values()])
        except:
            _LOGGER.debug("could not retreive sessions.")
            return 0

    @property
    def total_charge(self) -> float:
        try:
            return sum([h.charge for h in self.model.values()])
        except:
            _LOGGER.debug("could not retreive total charge")
            return 0.0

    async def async_unpack(self, incoming: dict) -> dict[int, Day] | None:
        if isinstance(incoming, dict) and len(incoming):
            model = {}
            for i in range(0, 7):
                if i in incoming.keys():
                    m_ret = Day(incoming[i]["sessions"], incoming[i]["total_charge"])
                else:
                    m_ret = Day(0, 0)
                model[i] = m_ret
            self.model = model
        else:
            return await self.async_set_init_model()

    async def async_update(self, _charge: float, mock_time: float | None = None):
        timer = mock_time or time.time()
        if _charge > 0:
            self.model[datetime.fromtimestamp(timer).weekday()].charge += _charge
            self.model[datetime.fromtimestamp(timer).weekday()].sessions += 1

    async def async_average_for_day(self, day: int) -> float:
        try:
            return self.model[day].charge / self.model[day].sessions
        except:
            return 0.0


# EXAMPLE = {
#     0: Day(1,50),
#     1: Day(4,50),
#     2: Day(2,15),
#     3: Day(3,20),
#     4: Day(10,123),
#     5: Day(0,0),
#     6: Day(10, 110)
# }

# EXPORT_EXAMPLE = {0: {'sessions': 1, 'total_charge': 50}, 1: {'sessions': 4, 'total_charge': 50}, 2: {'sessions': 2, 'total_charge': 15}, 3: {'sessions': 3, 'total_charge': 20}, 4: {'sessions': 10, 'total_charge': 123}, 5: {'sessions': 0, 'total_charge': 0}, 6: {'sessions': 10, 'total_charge': 110}}
# EXPORT_MISSING = {0: {'sessions': 1, 'total_charge': 50}, 2: {'sessions': 2, 'total_charge': 15}, 3: {'sessions': 3, 'total_charge': 20}, 4: {'sessions': 10, 'total_charge': 123}, 5: {'sessions': 0, 'total_charge': 0}, 6: {'sessions': 10, 'total_charge': 110}}
# e = EnergyWeekly()
# e.unpack(EXPORT_MISSING)
# print(e.export)
# print(e.model)
# print(e.average)
# print(e.average_for_day(0))



from statistics import mean
from ...models.locale.price.models.tiered_price import TieredPrice
from ...models.locale.enums.price_type import PriceType

def get_observed_peak(price_type: PriceType, peaks: list[float], tiers: list[TieredPrice]) -> float:
        if price_type != PriceType.Tiered:
            return mean(peaks)
        
        limits = [tier.upper_peak_limit for tier in tiers]
        suggest = min(l for l in limits if l > mean(peaks)) - 0.1

        if len(peaks) == 1:
            return round(suggest,1)

        if all(p < suggest for p in peaks):
            return round(suggest,1)

        replaced_set = peaks
        
        for _ in range(50):
            min_count = replaced_set.count(min(replaced_set))
            
            if min_count == len(replaced_set):
                """All peaks have normalized. The suggest will suffice"""
                return round(suggest,1)
            
            if min_count == len(replaced_set) - 1:
                """All peaks have normalized except one. We can calculate the remainder"""
                return _remainder_min(suggest, replaced_set)
            
            if replaced_set.count(max(replaced_set)) == len(replaced_set) - 1:
                """All peaks have normalized except one. We can calculate the remainder"""
                return _remainder_max(suggest, replaced_set)
            
            replaced_set = _count_up_min(suggest, replaced_set)
        return round(min(replaced_set),1)

def _remainder_max(suggest: float, values: list[float]) -> float:
    remaining_difference = suggest * len(values) - sum(values)
    values = [round(p + remaining_difference, 2) if p < max(values) else p for p in values]
    return round(min(values),1)

def _remainder_min(suggest: float, values: list[float]) -> float:
    needed_addition = (suggest * len(values) - sum(values)) / len(values)
    values = [min(round(p + needed_addition, 2), suggest) if p < suggest else p for p in values]
    remaining_difference = suggest * len(values) - sum(values)
    values = [min(round(p + remaining_difference / values.count(min(values)), 2), suggest) if p == min(values) else p for p in values]
    return round(min(values),1)

def _count_up_min(suggest: float, values: list[float]) -> list:
    min_value = min(values)
    second_min_value = min(h for h in values if h > min_value)
    current_mean = sum(values) / len(values)
    min_count = values.count(min_value)

    addition = min(
        second_min_value - min_value,
        max(0.1, (suggest - current_mean) / min_count)
    )
    values = [round(p+addition,2) if p == min(values) else p for p in values]
    return values


from __future__ import annotations
import logging
from typing import Tuple
from ...util import nametoid

_LOGGER = logging.getLogger(__name__)


class HubMember:
    def __init__(
        self,
        data_type,
        listenerentity=None,
        initval=None,
        name=None,
        init_override: bool = False,
        hass=None,
    ):
        self._value = initval
        self._type = data_type
        self._listenerentity, self._listenerattribute = self._set_listeners(
            listenerentity
        )
        self.name = name
        self.id = nametoid(self.name) if self.name is not None else None
        self.hass = hass
        self.warned_not_initialized = False
        self._is_initialized = init_override

    def _set_listeners(self, listenerentity: str | None) -> Tuple[str, str | None]:
        """Sets the listening entity. If it contains a |, the value after that will be considered the listener-attribute."""
        if listenerentity is None:
            return None, None
        try:
            _arr = listenerentity.split("|")
            if len(_arr) == 1:
                return listenerentity, None
            return _arr[0], _arr[1]
        except:
            _LOGGER.debug(
                f"could not handle this listenerentity in split-setup: {listenerentity}"
            )
            return None, None

    def _get_listeners(self) -> str:
        if self._listenerattribute is None:
            return self._listenerentity
        return f"{self._listenerentity}|{self._listenerattribute}"

    def update(self):
        try:
            self.value = self.get_sensor_from_hass(self._get_listeners())
        except:
            pass

    @property
    def is_initialized(self) -> bool:
        if self._is_initialized:
            return True
        if isinstance(self.value, self._type):
            _LOGGER.debug(f"{self._listenerentity} has initialized")
            self._is_initialized = True
            return True
        if not self.warned_not_initialized:
            _LOGGER.error(f"{self._listenerentity} was not {self._type}. {self.value}")
            self.warned_not_initialized = True
        return False

    @property
    def entity(self) -> str:
        return self._listenerentity

    @entity.setter
    def entity(self, val: str):
        self._listenerentity = val

    @property
    def use_attribute(self) -> bool:
        return self._listenerattribute is not None

    @property
    def attribute(self) -> str:
        return self._listenerattribute or ""

    @property
    def value(self):
        if self.use_attribute:
            self.update()
        return self._value

    @value.setter
    def value(self, value):
        self._value = self._set_value(value)

    def _set_value(self, value, passthrough: bool = False):
        if self._listenerattribute is not None and not passthrough:
            return self._set_value(
                self.get_sensor_from_hass(self._get_listeners()), True
            )
        if self._type is str:
            return str(value).lower()
        elif isinstance(value, self._type):
            return value
        elif self._type is float:
            try:
                return float(value)
            except ValueError:
                return 0
        elif self._type is int:
            try:
                return int(float(value))
            except ValueError:
                return 0
        elif self._type is bool:
            try:
                if value is None:
                    return False
                elif value.lower() == "on":
                    return True
                elif value.lower() == "off":
                    return False
            except ValueError as e:
                msg = f"Could not parse bool, setting to false to be sure {value}, {self._listenerentity}, {e}"
                _LOGGER.error(msg)
                return False

    def get_sensor_from_hass(self, sensor: str):
        if self.hass is not None:
            _sensor = sensor.split("|")
            if len(_sensor) == 2:
                ret = self.hass.states.get(_sensor[0])
                if ret:
                    if _sensor[1] is not None:
                        ret_attr = ret.attributes.get(_sensor[1])
                        return ret_attr
                    return ret
            elif len(_sensor) == 1:
                ret = self.hass.states.get(_sensor[0])
                if ret:
                    return ret
                else:
                    _LOGGER.warning(f"no state found for sensor: {_sensor[0]}")
        raise ValueError(f"no hass or no sensor found for {sensor}")

    async def async_get_sensor_from_hass(self, sensor: str):
        return self.get_sensor_from_hass(sensor)



from dataclasses import dataclass, field
from typing import List
#from .hours_model import HoursModel
from .hourselection_options import HourSelectionOptions
import logging

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=False)
class HourSelectionModel:
    adjusted_average: float | None = None
    current_peak: float = 0.0
    options: HourSelectionOptions = field(
        default_factory=lambda: HourSelectionOptions()
    )

    def __post_init__(self):
        self.validate()

    def validate(self):
        if isinstance(self.adjusted_average, (int, float)):
            assert self.adjusted_average >= 0
        else:
            assert self.adjusted_average is None



from datetime import datetime, date, time
import math
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .schedule_session import ScheduleSession
import logging

_LOGGER = logging.getLogger(__name__)


class Scheduler:
    """This class obj is what constitutes a running scheduler."""

    def __init__(self, options: HourSelectionOptions | None = None, test: bool = False):
        self.model = ScheduleSession(hourselection_options=options)
        self.active = False
        self.is_test = test

    @property
    def scheduler_active(self) -> bool:
        if self.active is False:
            return False
        return (
            self.model.departuretime > datetime.now() or self.model.remaining_charge > 0
        )

    async def async_create(
        self,
        desired_charge: float,
        departuretime: datetime,
        starttime: datetime = datetime.now(),
        override_settings=False,
    ):
        if not await self.async_check_parameters(
            desired_charge, departuretime, starttime
        ):
            return
        if self.scheduler_active:
            await self.async_cancel()
        self.model.departuretime = departuretime
        self.model.starttime = starttime
        self.model.remaining_charge = desired_charge
        self.model._override_settings = override_settings

    async def async_check_parameters(
        self, desired_charge: float, departuretime: datetime, starttime: datetime
    ) -> bool:
        if departuretime < starttime:
            _LOGGER.error("Starttime must be before departuretime if added.")
            return False
        if desired_charge <= 0:
            _LOGGER.error("Desired charge for scheduler must be greater than 0")
            return False
        if departuretime <= datetime.now() and self.is_test is False:
            _LOGGER.error(
                f"Departuretime must be in the future. You added: {departuretime}"
            )
            return False
        return True

    async def async_update(
        self,
        avg24: float,
        peak: float,
        charged_amount: float | None = None,
        prices: list | None = None,
        prices_tomorrow: list | None = None,
        mockdt: datetime | None = None,
    ):
        """calculate based on the pricing of hours, current peak and the avg24hr energy consumption"""
        self.model._mock_dt = mockdt if mockdt is not None else datetime.now()
        self.active = True
        if any(
            [
                self.model.remaining_charge <= 0,
                self.model.departuretime <= self.model._mock_dt,
            ]
        ):
            return await self.async_cancel()
        charge_per_hour = peak - (avg24 / 1000)
        if charge_per_hour <= 0:
            raise Exception

        self.model.remaining_charge -= (
            charged_amount if charged_amount is not None else 0
        )
        self.model.hours_price = [prices, prices_tomorrow]
        cheapest = await self.async_sort_pricelist()
        self.model.hours_charge = await self.async_get_charge_hours(
            cheapest_hours=cheapest, charge_per_hour=charge_per_hour, peak=peak
        )

    async def async_cancel(self):
        self.active = False
        self.model.departuretime = datetime.min
        self.model.starttime = datetime.min
        self.model.remaining_charge = 0

    async def async_sort_pricelist(self) -> dict:
        if (
            self.model._override_settings is False
            and self.model.hourselection_options is not None
        ):
            return await self.async_filter_pricelist()
        return dict(sorted(self.model.hours_price.items(), key=lambda item: item[1]))

    async def async_filter_pricelist(self) -> dict:
        filtered = {
            key: value
            for (key, value) in self.model.hours_price.items()
            if value <= self.model.hourselection_options.absolute_top_price
        }
        ret = dict(sorted(filtered.items(), key=lambda item: item[1]))
        return ret

    async def async_get_charge_hours(
        self, cheapest_hours: dict, charge_per_hour: float, peak: float
    ) -> dict:
        remainder = self.model.remaining_charge
        chargehours = dict()
        for c in cheapest_hours.keys():
            if remainder <= 0:
                break
            if remainder > charge_per_hour:
                chargehours[c] = 1
            elif 0 < remainder < charge_per_hour:
                chargehours[c] = math.ceil((remainder / peak) * 10) / 10
            remainder -= charge_per_hour
        return chargehours



import logging

from peaqevcore.common.spotprice.const import *
from peaqevcore.common.spotprice.energidataservice import EnergiDataServiceUpdater
from peaqevcore.common.spotprice.nordpool import NordPoolUpdater
from peaqevcore.common.spotprice.spotpricebase import SpotPriceBase
from peaqevcore.common.models.peaq_system import PeaqSystem

_LOGGER = logging.getLogger(__name__)


from enum import Enum

class SpotPriceType(Enum):
    NordPool = NORDPOOL
    EnergidataService = ENERGIDATASERVICE


class SpotPriceFactory:

    sources = {
        SpotPriceType.NordPool: NordPoolUpdater,
        SpotPriceType.EnergidataService: EnergiDataServiceUpdater
    }

    @staticmethod
    def create(
        hub, 
        observer, 
        system: PeaqSystem,
        test:bool = False, 
        is_active: bool = False
        ) -> SpotPriceBase:
        if test:
            return NordPoolUpdater(hub, test, system, observer)
        source = SpotPriceFactory.test_connections(hub.state_machine)
        return SpotPriceFactory.sources[source](hub, observer, system, test, is_active)

    @staticmethod
    def test_connections(hass) -> SpotPriceType:
        sensor = hass.states.get(ENERGIDATASERVICE_SENSOR)       
        if sensor:
            _LOGGER.debug("Found sensor %s", sensor)
            return SpotPriceType.EnergidataService
        else:
            _LOGGER.debug("No sensor %s", sensor)
            return SpotPriceType.NordPool
                

    


import asyncio
import logging

from peaqevcore.common.models.observer_types import ObserverTypes
from peaqevcore.common.spotprice.const import (ENERGIDATASERVICE, ENERGIDATASERVICE_SENSOR)
from peaqevcore.common.spotprice.models.spotprice_dto import EnergiDataServiceDTO
from peaqevcore.common.spotprice.spotpricebase import SpotPriceBase
from peaqevcore.common.models.peaq_system import PeaqSystem

_LOGGER = logging.getLogger(__name__)


class EnergiDataServiceUpdater(SpotPriceBase):
    def __init__(self, hub, observer, system:PeaqSystem, test:bool = False, is_active: bool = True):
        super().__init__(hub=hub, source=ENERGIDATASERVICE, system=system, test=test, is_active=is_active, observer=observer)

    async def async_set_dto(self, ret, initial: bool = False) -> None:
        _result = EnergiDataServiceDTO()
        await _result.set_model(ret)
        if await self.async_update_set_prices(_result):
            if initial:
                await self.hub.async_update_prices(
                    [self.model.prices, self.model.prices_tomorrow]
                )
                await self.observer.async_broadcast(ObserverTypes.SpotpriceInitialized)
            else:
                await self.observer.async_broadcast(
                    ObserverTypes.PricesChanged,
                    [self.model.prices, self.model.prices_tomorrow],
                )
            self._is_initialized = True

    def setup(self):
        try:
            sensor = self.state_machine.states.get(ENERGIDATASERVICE_SENSOR)
            if not sensor.state:
                self.hub.options.price.price_aware = False  # todo: composition
                _LOGGER.error(
                    f"There were no Spotprice-entities. Cannot continue. with price-awareness."
                )
            else:
                self.model.entity = ENERGIDATASERVICE_SENSOR
                _LOGGER.debug(
                    f"EnergiDataService has been set up and is ready to be used with {self.model.entity}"
                )
                asyncio.run_coroutine_threadsafe(
                    self.async_update_spotprice(initial=True),
                    self.state_machine.loop,
                )
        except Exception as e:
            self.hub.options.price.price_aware = False  # todo: composition
            _LOGGER.error(
                f"I was unable to get a Spotprice-entity. Cannot continue. with price-awareness: {e}"
            )


import logging
from typing import Tuple
from .hoursbase import Hours
from ..hoursselection import Hoursselection as core_hours
from ....common.enums.cautionhourtype import CautionHourType
from ...timer.timer import Timer
from ...scheduler.scheduler_facade import SchedulerFacade

_LOGGER = logging.getLogger(__name__)


class PriceAwareHours(Hours):
    def __init__(self, hub):
        self._hub = hub
        self.timer = Timer()
        self._cautionhour_type_string = hub.options.price.cautionhour_type
        self._core = core_hours(
            absolute_top_price=self._set_absolute_top_price(
                hub.options.price.top_price
            ),
            min_price=hub.options.price.min_price,
            cautionhour_type=self._cautionhour_type_string,
            non_hours=hub.options.nonhours
        )
        self._hass = hub.state_machine
        self._prices = []
        self.scheduler = SchedulerFacade(hub=self._hub, options=self.options)
        super().__init__(price_aware=True)

    @property
    def options(self):
        return self._core.model.options

    @property
    def dynamic_caution_hours(self) -> dict:
        if self.scheduler.scheduler_active:
            return self.scheduler.caution_hours
        return self._core.dynamic_caution_hours

    @property
    def cautionhour_type_string(self) -> str:
        return self._cautionhour_type_string

    @property
    def non_hours(self) -> list:
        if self.scheduler.scheduler_active:
            return self.scheduler.non_hours
        return self._core.non_hours

    @property
    def caution_hours(self) -> list:
        return self._core.caution_hours

    @property
    def future_hours(self) -> list:
        return self._core.future_hours

    @property
    def absolute_top_price(self):
        return self._core.model.options.absolute_top_price

    @property
    def min_price(self):
        return self._core.model.options.min_price

    @property
    def prices(self) -> list:
        return self._core.prices

    @property
    def prices_tomorrow(self) -> list:
        return self._core.prices_tomorrow

    @property
    def adjusted_average(self) -> float|None:
        return self._core.adjusted_average

    @adjusted_average.setter
    def adjusted_average(self, val):
        if isinstance(val, (float, int)):
            self._core.adjusted_average = val

    @property
    def stopped_string(self) -> str:
        return self._core.service.allowance.display_name

    @property
    def offsets(self) -> dict:
        return self._core.offsets

    @property
    def is_initialized(self) -> bool:
        if self.prices is not None:
            if len(self.prices):
                if self._is_initialized is False:
                    self._is_initialized = True
                    _LOGGER.debug("Hourselection has initialized")
                return True
        return False

    async def async_update_top_price(self, dyn_top_price) -> None:
        if self._hub.options.price.dynamic_top_price:
            await self._core.async_update_top_price(dyn_top_price)

    async def async_update_prices(
        self, prices: list = [], prices_tomorrow: list = []
    ) -> None:
        await self._core.async_update_prices(prices, prices_tomorrow)

    async def async_update_adjusted_average(self, val):
        await self._core.async_update_adjusted_average(val)

    async def async_get_average_kwh_price(self) -> Tuple[float | None, float | None]:
        if self._is_initialized:
            try:
                return await self._core.async_get_average_kwh_price()
            except ZeroDivisionError as e:
                _LOGGER.warning(f"get_average_kwh_price could not be calculated: {e}")
            return 0, None
        _LOGGER.debug("get avg kwh price, not initialized")
        return 0, None

    async def async_get_total_charge(self) -> Tuple[float, float | None]:
        if self._is_initialized:
            try:
                return await self._core.async_get_total_charge(
                    self._hub.sensors.current_peak.value
                )
            except ZeroDivisionError as e:
                _LOGGER.warning(f"get_total_charge could not be calculated: {e}")
            return 0, None
        _LOGGER.debug("get avg kwh price, not initialized")
        return 0, None

    @staticmethod
    def _set_absolute_top_price(_input) -> float:
        if _input is None or _input <= 0:
            return float("inf")
        return _input

    async def async_update_max_min(
        self,
        max_charge: float,
        session_energy: float | None = None,
        car_connected: bool = False,
        limiter: float = 0.0
    ):
        if not self._core.service.max_min.active:
            await self._core.service.max_min.async_setup(max_charge)
        await self._core.service.max_min.async_update(
            avg24=self._hub.sensors.powersensormovingaverage24.value,
            peak=self._hub.sensors.current_peak.value,
            max_desired=max_charge,
            session_energy=session_energy,
            car_connected=car_connected,
            limiter=limiter
        )
